# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['xformula',
 'xformula.arch',
 'xformula.arch.bisect',
 'xformula.arch.enumeration',
 'xformula.arch.meta',
 'xformula.arch.runtime',
 'xformula.arch.runtime.reflection',
 'xformula.runtime',
 'xformula.runtime.core',
 'xformula.runtime.core.context',
 'xformula.runtime.core.context.abc',
 'xformula.syntax',
 'xformula.syntax.ast',
 'xformula.syntax.ast.nodes',
 'xformula.syntax.ast.nodes.abc',
 'xformula.syntax.core',
 'xformula.syntax.core.context',
 'xformula.syntax.core.context.abc',
 'xformula.syntax.core.customization',
 'xformula.syntax.core.customization.tagging',
 'xformula.syntax.core.features',
 'xformula.syntax.core.features.abc',
 'xformula.syntax.core.features.attributes',
 'xformula.syntax.core.features.attributes.ast',
 'xformula.syntax.core.features.attributes.ast.nodes',
 'xformula.syntax.core.features.attributes.definitions',
 'xformula.syntax.core.features.attributes.definitions.non_terminals',
 'xformula.syntax.core.features.base',
 'xformula.syntax.core.features.base.definitions',
 'xformula.syntax.core.features.base.definitions.directives',
 'xformula.syntax.core.features.base.definitions.terminals',
 'xformula.syntax.core.features.calls',
 'xformula.syntax.core.features.calls.ast',
 'xformula.syntax.core.features.calls.ast.nodes',
 'xformula.syntax.core.features.calls.definitions',
 'xformula.syntax.core.features.calls.definitions.non_terminals',
 'xformula.syntax.core.features.containers',
 'xformula.syntax.core.features.containers.ast',
 'xformula.syntax.core.features.containers.ast.nodes',
 'xformula.syntax.core.features.containers.definitions',
 'xformula.syntax.core.features.containers.definitions.non_terminals',
 'xformula.syntax.core.features.expressions',
 'xformula.syntax.core.features.expressions.definitions',
 'xformula.syntax.core.features.expressions.definitions.non_terminals',
 'xformula.syntax.core.features.identifiers',
 'xformula.syntax.core.features.identifiers.ast',
 'xformula.syntax.core.features.identifiers.ast.nodes',
 'xformula.syntax.core.features.identifiers.definitions',
 'xformula.syntax.core.features.identifiers.definitions.non_terminals',
 'xformula.syntax.core.features.identifiers.definitions.terminals',
 'xformula.syntax.core.features.literals',
 'xformula.syntax.core.features.literals.ast',
 'xformula.syntax.core.features.literals.ast.nodes',
 'xformula.syntax.core.features.literals.definitions',
 'xformula.syntax.core.features.literals.definitions.non_terminals',
 'xformula.syntax.core.features.literals.definitions.terminals',
 'xformula.syntax.core.features.mappings',
 'xformula.syntax.core.features.mappings.ast',
 'xformula.syntax.core.features.mappings.ast.nodes',
 'xformula.syntax.core.features.mappings.definitions',
 'xformula.syntax.core.features.mappings.definitions.non_terminals',
 'xformula.syntax.core.features.operations',
 'xformula.syntax.core.features.operations.ast',
 'xformula.syntax.core.features.operations.ast.nodes',
 'xformula.syntax.core.features.operations.definitions',
 'xformula.syntax.core.features.operations.definitions.non_terminals',
 'xformula.syntax.core.features.operations.generators',
 'xformula.syntax.core.features.operations.runtime',
 'xformula.syntax.core.features.operations.runtime.reflection',
 'xformula.syntax.core.features.polyfill',
 'xformula.syntax.core.features.polyfill.definitions',
 'xformula.syntax.core.features.polyfill.generators',
 'xformula.syntax.core.features.polyfill.runtime',
 'xformula.syntax.core.features.polyfill.runtime.reflection',
 'xformula.syntax.core.features.primaries',
 'xformula.syntax.core.features.primaries.definitions',
 'xformula.syntax.core.features.primaries.definitions.non_terminals',
 'xformula.syntax.core.features.separators',
 'xformula.syntax.core.features.separators.definitions',
 'xformula.syntax.core.features.separators.definitions.templates',
 'xformula.syntax.core.features.terms',
 'xformula.syntax.core.features.terms.definitions',
 'xformula.syntax.core.features.terms.definitions.non_terminals',
 'xformula.syntax.core.features.validation',
 'xformula.syntax.core.features.validation.detectors',
 'xformula.syntax.core.features.validation.errors',
 'xformula.syntax.core.features.wrappers',
 'xformula.syntax.core.features.wrappers.definitions',
 'xformula.syntax.core.features.wrappers.definitions.templates',
 'xformula.syntax.core.operations',
 'xformula.syntax.core.operations.abc',
 'xformula.syntax.grammar',
 'xformula.syntax.grammar.definitions',
 'xformula.syntax.grammar.definitions.abc',
 'xformula.syntax.grammar.definitions.runtime',
 'xformula.syntax.grammar.definitions.runtime.reflection',
 'xformula.syntax.grammar.directives',
 'xformula.syntax.grammar.directives.abc',
 'xformula.syntax.grammar.ebnf',
 'xformula.syntax.grammar.non_terminals',
 'xformula.syntax.grammar.non_terminals.abc',
 'xformula.syntax.grammar.regex',
 'xformula.syntax.grammar.templates',
 'xformula.syntax.grammar.templates.abc',
 'xformula.syntax.grammar.terminals',
 'xformula.syntax.grammar.terminals.abc',
 'xformula.syntax.lexer',
 'xformula.syntax.lexer.tokens',
 'xformula.syntax.lexer.tokens.abc',
 'xformula.syntax.parser',
 'xformula.syntax.parser.transformers',
 'xformula.syntax.parser.transformers.abc',
 'xformula.syntax.parser.transformers.runtime',
 'xformula.syntax.parser.transformers.runtime.reflection',
 'xformula.syntax.parser.trees',
 'xformula.syntax.parser.trees.abc']

package_data = \
{'': ['*']}

install_requires = \
['lark>=1.1.4']

setup_kwargs = {
    'name': 'xformula',
    'version': '0.1.1',
    'description': 'Highly customizable language front-end, aimed to be a base for custom DSL evaluators.',
    'long_description': "# XFormula\n\nHighly customizable language front-end, aimed to be a base\nfor custom DSL evaluators.\n\n---\n\n**Developer note:** I've created this library to use in many of my projects and\ndecided to publish the source code, because it may help somebody to write a DSL\ncompiler or evaluator can get benefit from Python ecosystem, within a reasonable\ntime.\n\nI couldn't write documentation because of my tight schedule in these days. \nBut I see the code as self-explanatory, feel free to read it if you're interested.\n\nIn the meantime, please note that **this project is still in development**.\n\n---\n\n\n## Features:\n\n- Built on top of [Lark Parser Toolkit](https://lark-parser.readthedocs.io/en/latest/)\n- - LALR(1), Earley and CYK parsing algorithms are supported by Lark;\n    XFormula uses LALR(1) by default\n- Automatic EBNF grammar generator via declarative Python functions\n  (The final grammar generated using the default features can be found in\n  [out/Grammar.lark](https://github.com/ertgl/xformula/blob/main/out/Grammar.lark)\n  file)\n- Modular featurization system to manipulate grammar and parser context dynamically. \n- A ready-to-use compact dialect that supports some general purpose data types\n  and basic symbols\n  (See `xformula.syntax.ast` package)\n\nSee `xformula.syntax.core.features` package and\n`xformula.syntax.core.operations.default_operator_precedences` module for more\ninformation about the default behaviours.\n\n\n## License\n\n[MIT](https://github.com/ertgl/xformula/blob/main/LICENSE)\n",
    'author': 'Ertuğrul Keremoğlu',
    'author_email': 'ertugkeremoglu@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



from enum import IntEnum

from xformula.arch.enumeration import Iota

__all__ = [
    "Associativity",
]


__xf__associativity_iota__ = Iota()


class Associativity(IntEnum):

    UNDEFINED = __xf__associativity_iota__()

    __XF__START__ = __xf__associativity_iota__()

    LEFT_TO_RIGHT = __xf__associativity_iota__()

    RIGHT_TO_LEFT = __xf__associativity_iota__()

    __XF__END__ = __xf__associativity_iota__()


del __xf__associativity_iota__



from abc import ABC

from xformula.syntax.parser.transformers.abc.transformer import Transformer

__all__ = [
    "ASTBuilderProtocol",
]


class ASTBuilderProtocol(
    Transformer,
    ABC,
):
    ...



from typing import TypeVar

from xformula.syntax.grammar.templates.abc import Template

__all__ = [
    "SeparatedPair",
]


T = TypeVar("T")


class SeparatedPair(
    Template[list[T]],
):
    class Meta:

        arguments = [
            Template.arg("sep"),
            Template.arg("lhs"),
            Template.arg("rhs"),
        ]

    def build_grammar(self) -> str:
        define = self.ebnf.define
        arg = self.ebnf.arg

        return define(arg("lhs"), arg("sep"), arg("rhs"))



from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.operand import Operand

__all__ = [
    "Primary",
]


class Primary(
    Operand,
    Node,
):
    class Meta(
        Operand.Meta,
        Node.Meta,
    ):

        abstract = True

        is_primary = True



from xformula.syntax.parser.parser import Parser
from xformula.syntax.parser.transformers import ASTBuilderProtocol, Transformer
from xformula.syntax.parser.trees import ParseTree

__all__ = [
    "ASTBuilderProtocol",
    "ParseTree",
    "Parser",
    "Transformer",
]



from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.terminals.abc import Terminal
from xformula.syntax.lexer.tokens.abc import Token

__all__ = [
    "IDENTIFIER",
]


class IDENTIFIER(
    Terminal[str],
):
    class Meta:

        tags = {
            non_terminal("Identifier"): 2000,
        }

    def build_grammar(self) -> str:
        define = self.ebnf.define
        regex = self.ebnf.regex

        return define(regex(r"(?<!\d)[^\W0-9]\w*"))

    def transform_token(
        self,
        runtime_context: RuntimeContext,
        token: Token,
    ) -> str:
        return token.value



__all__ = [
    "optional",
]


def optional(*expressions: str) -> str:
    inner = " ".join(expressions)
    wrap = len(expressions) > 1 or " " in inner
    if wrap:
        return f"({inner})?"
    return f"{inner}?"



from xformula.syntax.grammar.ebnf._directive import directive
from xformula.syntax.grammar.ebnf._flat_join import flat_join

__all__ = [
    "use_directive",
]


def use_directive(
    name: str,
    *expressions: str,
) -> str:
    return f"%{directive(name)} {flat_join(' ', expressions)}"



from xformula.syntax.core.features.abc.feature import Feature
from xformula.syntax.core.features.abc.feature_meta import FeatureMeta
from xformula.syntax.core.features.abc.feature_options import FeatureOptions

__all__ = [
    "Feature",
    "FeatureMeta",
    "FeatureOptions",
]



import string

__all__ = [
    "escape_char",
]


def escape_char(char: str) -> str:
    if len(char) > 1:
        return char
    if char.isspace():
        return r"(\s+)"
    if char in string.punctuation:
        return rf"\{char}"
    return char



from xformula.arch.meta import Configurable
from xformula.syntax.ast.nodes.abc.node_meta import NodeMeta

__all__ = [
    "Node",
]


class Node(
    Configurable,
    metaclass=NodeMeta,
):
    class Meta(Configurable.Meta):

        abstract: bool = True

        is_start_point: bool = False

        is_expression: bool = False

        is_simple_expression: bool = False

        is_operation: bool = False

        is_operator: bool = False

        is_symbol: bool = True

        is_operand: bool = False

        is_primary: bool = False

        is_mapping: bool = False

        is_pair: bool = False

        is_container: bool = False

        is_term: bool = False

        is_call: bool = False

        is_attribute: bool = False

        is_identifier: bool = False

        is_literal: bool = False

        has_associativity_attribute: bool = False

        has_arity_attribute: bool = False

        has_arguments_attribute: bool = False

        has_callee_attribute: bool = False

        has_context_attribute: bool = False

        has_elements_attribute: bool = False

        has_name_attribute: bool = False

        has_owner_attribute: bool = False

        has_placement_attribute: bool = False

        has_precedence_level_attribute: bool = False

        has_type_attribute: bool = False

        has_operands_attribute: bool = False

        has_operator_attribute: bool = False

        has_symbols_attribute: bool = False

        has_value_attribute: bool = False



from xformula.syntax.grammar.definitions import (
    Definition,
    DefinitionMeta,
    DefinitionOptions,
    DefinitionType,
    EBNFExpressionBuilderProtocol,
    RegexExpressionBuilderProtocol,
)
from xformula.syntax.grammar.directives import Directive
from xformula.syntax.grammar.non_terminals import NonTerminal
from xformula.syntax.grammar.templates import Template
from xformula.syntax.grammar.terminals import Terminal

__all__ = [
    "Definition",
    "DefinitionMeta",
    "DefinitionOptions",
    "DefinitionType",
    "Directive",
    "EBNFExpressionBuilderProtocol",
    "NonTerminal",
    "RegexExpressionBuilderProtocol",
    "Template",
    "Terminal",
]



from xformula.syntax.core.features.containers.definitions.non_terminals.list import List
from xformula.syntax.core.features.containers.definitions.non_terminals.set import Set
from xformula.syntax.core.features.containers.definitions.non_terminals.tuple import (
    Tuple,
)

__all__ = [
    "List",
    "Set",
    "Tuple",
]



import re
from typing import cast

from lark import Lark

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.ast.nodes.abc import Node
from xformula.syntax.core.context import SyntaxContext
from xformula.syntax.core.features import DEFAULT_FEATURE_TYPES, Feature
from xformula.syntax.core.operations import (
    DEFAULT_OPERATOR_PRECEDENCES,
    OperatorPrecedence,
)
from xformula.syntax.grammar.ebnf import document
from xformula.syntax.parser.transformers import ASTBuilderProtocol

__all__ = [
    "Parser",
]


DEBUG = False


class Parser:

    _ebnf_document: str

    _grammar: Lark

    ast_builder: ASTBuilderProtocol

    syntax_context: SyntaxContext

    runtime_context: RuntimeContext

    @classmethod
    def get_default_feature_types(cls) -> list[type[Feature]]:
        return DEFAULT_FEATURE_TYPES

    @classmethod
    def get_default_operator_precedences(cls) -> list[OperatorPrecedence]:
        return DEFAULT_OPERATOR_PRECEDENCES

    def __init__(
        self,
        ast_builder: ASTBuilderProtocol | None = None,
        runtime_context: RuntimeContext | None = None,
        syntax_context: SyntaxContext | None = None,
    ) -> None:
        if runtime_context is None:
            runtime_context = RuntimeContext()
        self.runtime_context = runtime_context
        if syntax_context is None:
            syntax_context = SyntaxContext(
                feature_types=self.__class__.get_default_feature_types(),
                operator_precedences=self.__class__.get_default_operator_precedences(),
            )
        self.syntax_context = syntax_context
        self._ebnf_document = document(syntax_context)
        if DEBUG:
            print("*" * 100)
            print(self._ebnf_document)
            print("*" * 100)
            print()
        self._grammar = Lark(
            self._ebnf_document,
            lexer="contextual",
            parser="lalr",
            g_regex_flags=re.UNICODE | re.IGNORECASE,
            propagate_positions=True,
            maybe_placeholders=False,
        )
        if ast_builder is None:
            ast_builder_class = syntax_context.build_ast_builder_class(runtime_context)
            ast_builder = ast_builder_class()
        self.ast_builder = ast_builder

    @property
    def ebnf_document(self) -> str:
        return self._ebnf_document

    @property
    def grammar(self) -> Lark:
        return self._grammar

    def parse(
        self,
        source: str,
    ) -> Node:
        parse_tree = self.grammar.parse(source)
        ast = cast(Node, self.ast_builder.transform(parse_tree))
        return ast



from typing import TYPE_CHECKING

from xformula.syntax.grammar.ebnf._keyword import keyword
from xformula.syntax.grammar.ebnf._literal import literal
from xformula.syntax.grammar.ebnf._op import op

if TYPE_CHECKING:
    from xformula.syntax.ast.nodes.abc.symbol_type import SymbolType

__all__ = [
    "sym",
]


def sym(info: tuple["SymbolType", str]) -> str:
    from xformula.syntax.ast.nodes.abc import SymbolType

    if info[0] == SymbolType.KEYWORD:
        return keyword(info[1])

    if info[0] == SymbolType.OPERATOR:
        return op(info[1])

    return literal(info[1])



from typing import Any, Callable

from xformula.arch.meta import Configurable
from xformula.syntax.core.context import SyntaxContext
from xformula.syntax.grammar.definitions.abc.definition_meta import DefinitionMeta
from xformula.syntax.grammar.definitions.abc.definition_type import DefinitionType
from xformula.syntax.grammar.definitions.abc.ebnf_expression_builder_protocol import (
    EBNFExpressionBuilderProtocol,
)
from xformula.syntax.grammar.definitions.abc.regex_expression_builder_protocol import (
    RegexExpressionBuilderProtocol,
)

__all__ = [
    "Definition",
]


class Definition(
    Configurable,
    metaclass=DefinitionMeta,
):

    context: SyntaxContext

    ebnf: EBNFExpressionBuilderProtocol

    regex: RegexExpressionBuilderProtocol

    class Meta(Configurable.Meta):

        abstract: bool = True

        definition_name: str = ""

        atomic: bool = False

        retain_anonymous_literals: bool = False

        type: DefinitionType = DefinitionType.UNDEFINED

        arguments: list[str] = []

        priority: int = 0

        tags: dict[str, int | float] = dict()

        transform: str | Callable[..., Any] = "transform"

    def __init__(
        self,
        context: SyntaxContext,
    ) -> None:
        self.context = context
        self.ebnf = self.context.ebnf_expression_builder_class(self)
        self.regex = self.context.regex_expression_builder_class(self)

    def build_grammar(self) -> str:
        raise NotImplementedError(
            f"{self.__class__.__qualname__}.build_grammar is not implemented"
        )



from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.terminals.abc import Terminal
from xformula.syntax.lexer.tokens.abc import Token

__all__ = [
    "DECIMAL_INT",
]


class DECIMAL_INT(  # NOSONAR
    Terminal[int],
):
    class Meta:

        tags = {
            non_terminal("Int"): -4000,
        }

    def build_grammar(self) -> str:
        define = self.ebnf.define
        regex = self.ebnf.regex

        any_char_of = self.regex.any_char_of
        any_of = self.regex.any_of
        chars_between = self.regex.chars_between
        flat = self.regex.flat
        flat_optional = self.regex.flat_optional
        flat_zero_or_more_of = self.regex.flat_zero_or_more_of
        non_capturing_group = self.regex.non_capturing_group

        re_zero = flat(
            r"0",
            flat_zero_or_more_of(
                non_capturing_group(
                    flat_optional(r"_"),
                    "0",
                ),
            ),
        )

        re_non_zero = flat(
            any_char_of(chars_between(r"1", r"9")),
            flat_zero_or_more_of(
                non_capturing_group(
                    flat_optional(r"_"),
                    any_char_of(chars_between(r"0", r"9")),
                ),
            ),
        )

        return define(regex(any_of(re_zero, re_non_zero)))

    def transform_token(
        self,
        runtime_context: RuntimeContext,
        token: Token,
    ) -> int:
        return int(token.value, base=10)



import dataclasses

from xformula.syntax.ast.nodes.abc import Call as _Call
from xformula.syntax.ast.nodes.abc import Node

__all__ = [
    "Call",
]


@dataclasses.dataclass()
class Call(_Call):

    callee: Node = dataclasses.field(
        kw_only=True,
        default_factory=Node,
    )

    arguments: list[Node] = dataclasses.field(
        kw_only=True,
        default_factory=list,
    )



import dataclasses

from xformula.syntax.ast.nodes.abc import Pair, SimpleExpression

__all__ = [
    "CallKeywordArgument",
]


@dataclasses.dataclass()
class CallKeywordArgument(
    Pair[SimpleExpression, SimpleExpression],
):

    elements: tuple[SimpleExpression, SimpleExpression] = dataclasses.field(
        kw_only=True,
        default_factory=lambda: (SimpleExpression(), SimpleExpression()),
    )



import dataclasses

from xformula.syntax.ast.nodes import Literal

__all__ = [
    "Str",
]


@dataclasses.dataclass()
class Str(
    Literal[str],
):

    value: str = dataclasses.field(
        kw_only=True,
        default=str(),
    )



from xformula.syntax.core.features.abc import Feature

__all__ = [
    "PrimaryFeature",
]


class PrimaryFeature(Feature):
    class Meta:
        fqn = "xformula.core.Primary"

    def setup(self) -> None:
        from xformula.syntax.core.features.primaries.definitions.non_terminals import (
            Primary,
        )

        self.non_terminal_types.append(Primary)



from xformula.syntax.ast.nodes.abc.has_context import HasContext
from xformula.syntax.ast.nodes.abc.has_name import HasName
from xformula.syntax.ast.nodes.abc.has_owner import HasOwner
from xformula.syntax.ast.nodes.abc.identifier import Identifier
from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.term import Term

__all__ = [
    "Attribute",
]


class Attribute(
    Term,
    HasOwner[Node],
    HasName[Identifier],
    HasContext,
    Node,
):
    class Meta(
        Term.Meta,
        HasOwner.Meta,
        HasName.Meta,
        HasContext.Meta,
        Node.Meta,
    ):

        abstract = True

        is_attribute = True



from xformula.syntax.core.features.abc import Feature

__all__ = [
    "MappingFeature",
]


class MappingFeature(Feature):
    class Meta:
        fqn = "xformula.core.Mapping"

    def setup(self) -> None:
        from xformula.syntax.core.features.mappings.definitions.non_terminals import (
            Dict,
            DictPair,
        )

        self.non_terminal_types.extend(
            [
                DictPair,
                Dict,
            ],
        )



from typing import TypeVar

from xformula.syntax.ast.nodes.abc.container import Container
from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.pair import Pair

__all__ = [
    "Mapping",
]


K = TypeVar("K")

V = TypeVar("V")


class Mapping(
    Container[Pair[K, V]],
    Node,
):
    class Meta(
        Container.Meta,
        Node.Meta,
    ):

        abstract = True

        is_mapping = True



from typing import TYPE_CHECKING, Union

from xformula.syntax.grammar.ebnf._normalize_name_by_type import normalize_name_by_type

if TYPE_CHECKING:
    from xformula.syntax.grammar.definitions.abc.definition import Definition

__all__ = [
    "get_normalized_name_by_definition",
]


def get_normalized_name_by_definition(
    definition: Union[
        "Definition",
        type["Definition"],
    ],
) -> str:
    from xformula.syntax.grammar.definitions.abc import Definition

    definition_type = (
        definition.__class__ if isinstance(definition, Definition) else definition
    )

    return normalize_name_by_type(
        definition_type.options.type,
        definition_type.options.definition_name,
    )



from xformula.syntax.core.features.operations.ast.nodes.operation import Operation
from xformula.syntax.core.features.operations.ast.nodes.operator import Operator
from xformula.syntax.core.features.operations.ast.nodes.symbol import Symbol

__all__ = [
    "Operation",
    "Operator",
    "Symbol",
]



from xformula.syntax.core.features.operations.runtime.reflection.non_terminal_operation_class_builder import (
    NonTerminalOperationClassBuilder,
)

__all__ = [
    "NonTerminalOperationClassBuilder",
]



from typing import TYPE_CHECKING

from xformula.syntax.grammar.ebnf._indent import indent

if TYPE_CHECKING:
    from xformula.syntax.grammar.definitions.abc.definition import Definition

__all__ = [
    "define_rule",
]


def define_rule(definition: "Definition") -> str:
    from xformula.syntax.grammar.definitions.abc import DefinitionType

    header = definition.__class__.options.definition_name
    extra_indent_width = 1

    if definition.__class__.options.retain_anonymous_literals:
        header = f"!{header}"

    if (
        definition.__class__.options.type == DefinitionType.NON_TERMINAL
        and not definition.__class__.options.atomic
    ):
        header = f"?{header}"

    if definition.__class__.options.type == DefinitionType.TEMPLATE:
        args = ", ".join(definition.__class__.options.arguments)
        specify_args = "".join(["{", args, "}"])
        header = f"{header}{specify_args}"

    if definition.__class__.options.priority != 0:
        header = f"{header}.{definition.__class__.options.priority}"

    indent_width = len(header) + extra_indent_width

    grammar = definition.build_grammar()

    lines = grammar.split("\n")
    first_body_line, rest_body_lines = lines[0], lines[1:]

    formatted_first_line = f"{header} : {first_body_line}"
    formatted_trailing_lines = indent(*rest_body_lines, width=indent_width)

    formatted_body_split = [formatted_first_line]
    if formatted_trailing_lines:
        formatted_body_split.append(formatted_trailing_lines)

    formatted_grammar = "\n".join(formatted_body_split)

    return formatted_grammar



import dataclasses

__all__ = [
    "Iota",
]


@dataclasses.dataclass()
class Iota:

    value: int = dataclasses.field(
        default=-1,
    )

    step: int = dataclasses.field(
        default=1,
    )

    def incr(self) -> int:
        self.value += self.step
        return self.value

    def reset(self) -> int:
        self.value = 0
        return self.value

    def __call__(
        self,
        reset: bool = False,
    ) -> int:
        if reset:
            return self.reset()
        return self.incr()



from xformula.syntax.core.features.mappings.definitions.non_terminals.dict import Dict
from xformula.syntax.core.features.mappings.definitions.non_terminals.dict_pair import (
    DictPair,
)

__all__ = [
    "Dict",
    "DictPair",
]



from xformula.syntax.core.features.containers.ast.nodes.list import List
from xformula.syntax.core.features.containers.ast.nodes.set import Set
from xformula.syntax.core.features.containers.ast.nodes.tuple import Tuple

__all__ = [
    "List",
    "Set",
    "Tuple",
]



from xformula.syntax.core.features.literals.definitions.terminals.binary_int import (
    BINARY_INT,
)
from xformula.syntax.core.features.literals.definitions.terminals.bool import BOOL
from xformula.syntax.core.features.literals.definitions.terminals.complex import COMPLEX
from xformula.syntax.core.features.literals.definitions.terminals.decimal_int import (
    DECIMAL_INT,
)
from xformula.syntax.core.features.literals.definitions.terminals.float import FLOAT
from xformula.syntax.core.features.literals.definitions.terminals.hexadecimal_int import (
    HEXADECIMAL_INT,
)
from xformula.syntax.core.features.literals.definitions.terminals.none import NONE
from xformula.syntax.core.features.literals.definitions.terminals.octal_int import (
    OCTAL_INT,
)
from xformula.syntax.core.features.literals.definitions.terminals.single_line_str import (
    SINGLE_LINE_STR,
)

__all__ = [
    "BINARY_INT",
    "BOOL",
    "COMPLEX",
    "DECIMAL_INT",
    "FLOAT",
    "HEXADECIMAL_INT",
    "NONE",
    "OCTAL_INT",
    "SINGLE_LINE_STR",
]



from xformula.syntax.ast.nodes.abc.has_arity import HasArity
from xformula.syntax.ast.nodes.abc.has_associativity import HasAssociativity
from xformula.syntax.ast.nodes.abc.has_placement import HasPlacement
from xformula.syntax.ast.nodes.abc.has_precedence_level import HasPrecedenceLevel
from xformula.syntax.ast.nodes.abc.has_symbols import HasSymbols
from xformula.syntax.ast.nodes.abc.node import Node

__all__ = [
    "Operator",
]


class Operator(
    HasArity,
    HasAssociativity,
    HasPlacement,
    HasPrecedenceLevel,
    HasSymbols,
    Node,
):
    class Meta(
        HasArity.Meta,
        HasAssociativity.Meta,
        HasPlacement.Meta,
        HasPrecedenceLevel.Meta,
        HasSymbols.Meta,
        Node.Meta,
    ):

        abstract = True

        is_operator = True



from xformula.syntax.core.features.calls.definitions.non_terminals.call import Call
from xformula.syntax.core.features.calls.definitions.non_terminals.call_argument import (
    CallArgument,
)
from xformula.syntax.core.features.calls.definitions.non_terminals.call_keyword_argument import (
    CallKeywordArgument,
)

__all__ = [
    "Call",
    "CallArgument",
    "CallKeywordArgument",
]



from typing import overload

from xformula.syntax.grammar.regex._flat_join import flat_join

__all__ = [
    "wrap",
]


@overload
def wrap(
    outer: str,
    inner: str,
) -> str:
    ...


@overload
def wrap(
    left: str,
    right: str,
    inner: str,
) -> str:
    ...


@overload
def wrap(
    left: str,
    right: str,
    *inner: str,
) -> str:
    ...


def wrap(*args):
    if len(args) == 2:
        return f"{args[0]}{args[1]}{args[0]}"
    if len(args) == 3:
        return f"{args[0]}{args[2]}{args[1]}"
    return f"{args[0]}{args[2]}{flat_join('', args[3:])}{args[1]}"



from typing import TYPE_CHECKING, Iterable

from xformula.syntax.core.context.abc import SyntaxContext

if TYPE_CHECKING:
    from xformula.syntax.grammar.definitions.abc.definition import Definition

__all__ = [
    "TaggedDefinitionIterator",
]


class TaggedDefinitionIterator:

    syntax_context: SyntaxContext

    tag: str

    @classmethod
    def iterate(
        cls,
        syntax_context: SyntaxContext,
        tag: str,
    ) -> Iterable["Definition"]:
        iterator = cls(syntax_context, tag)
        return iterator.find_tagged_definitions()

    def __init__(
        self,
        syntax_context: SyntaxContext,
        tag: str,
    ) -> None:
        self.syntax_context = syntax_context
        self.tag = tag

    def find_tagged_definitions(self) -> Iterable["Definition"]:
        alternation = self.syntax_context.tagged_definitions.get(self.tag, [])
        yield from alternation



import dataclasses

from xformula.syntax.ast.nodes.abc import List as _List
from xformula.syntax.ast.nodes.abc import SimpleExpression

__all__ = [
    "List",
]


@dataclasses.dataclass()
class List(
    _List[SimpleExpression],
):

    elements: list[SimpleExpression] = dataclasses.field(
        kw_only=True,
        default_factory=list,
    )
